@@ -1,6 +1,6 @@
 ##############################################################################
 #
-# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
+# Copyright (c) 2001, 2002 Zope Corporation and Contributors.
 # All Rights Reserved.
 #
 # This software is subject to the provisions of the Zope Public License,
@@ -21,13 +21,13 @@ import logging
 
 import traceback, time
 
-import ThreadedAsync
 from ZEO.zrpc import smac
 from ZEO.zrpc.error import ZRPCError, DisconnectedError
 from ZEO.zrpc.marshal import Marshaller, ServerMarshaller
 from ZEO.zrpc.trigger import trigger
 from ZEO.zrpc.log import short_repr, log
 from ZODB.loglevels import BLATHER, TRACE
+import ZODB.POSException
 
 REPLY = ".reply" # message name used for replies
 ASYNC = 1
@@ -42,29 +42,30 @@ client_map = {}
 client_trigger = trigger(client_map)
 client_logger = logging.getLogger('ZEO.zrpc.client_loop')
 client_exit_event = threading.Event()
-client_running = True
+client_running = False
 def client_exit():
     global client_running
-    client_running = False
-    client_trigger.pull_trigger()
-    client_exit_event.wait()
+    if client_running:
+        client_running = False
+        client_trigger.pull_trigger()
+        client_exit_event.wait(99)
 
 atexit.register(client_exit)
 
 def client_loop():
-    map = client_map
+    global client_running
+    client_running = True
+    client_exit_event.clear()
 
+    map = client_map
     read = asyncore.read
     write = asyncore.write
     _exception = asyncore._exception
     loop_failures = 0
-    client_exit_event.clear()
-    global client_running
-    client_running = True
-    
+
     while client_running and map:
         try:
-            
+
             # The next two lines intentionally don't use
             # iterators. Other threads can close dispatchers, causeing
             # the socket map to shrink.
@@ -87,7 +88,7 @@ def client_loop():
                             continue
                         if [fd for fd in w if fd not in map]:
                             continue
-                        
+
                     raise
                 else:
                     continue
@@ -132,7 +133,7 @@ def client_loop():
         except:
             if map:
                 try:
-                    client_logger.critical('The ZEO cient loop failed.',
+                    client_logger.critical('The ZEO client loop failed.',
                                            exc_info=sys.exc_info())
                 except:
                     pass
@@ -153,9 +154,19 @@ def client_loop():
 
     client_exit_event.set()
 
-client_thread = threading.Thread(target=client_loop)
-client_thread.setDaemon(True)
-client_thread.start()
+client_thread_lock = threading.Lock()
+client_thread = None
+def start_client_thread():
+    client_thread_lock.acquire()
+    try:
+        global client_thread
+        if client_thread is None:
+            client_thread = threading.Thread(target=client_loop, name=__name__)
+            client_thread.setDaemon(True)
+            client_thread.start()
+    finally:
+        client_thread_lock.release()
+
 #
 ##############################################################################
 
@@ -338,20 +349,24 @@ class Connection(smac.SizedMessageAsyncConnection, object):
     #             receiveBlobStart
     #             receiveBlobChunk
     #             receiveBlobStop
-    
-    # XXX add blob methods
+    #
+    # Z309 -- named after the ZODB release 3.9
+    #         New server methods:
+    #             restorea, iterator_start, iterator_next,
+    #             iterator_record_start, iterator_record_next,
+    #             iterator_gc
 
     # Protocol variables:
     # Our preferred protocol.
-    current_protocol = "Z308"
+    current_protocol = "Z309"
 
     # If we're a client, an exhaustive list of the server protocols we
     # can accept.
-    servers_we_can_talk_to = [current_protocol]
+    servers_we_can_talk_to = ["Z308", current_protocol]
 
     # If we're a server, an exhaustive list of the client protocols we
     # can accept.
-    clients_we_can_talk_to = ["Z200", "Z201", "Z303", current_protocol]
+    clients_we_can_talk_to = ["Z200", "Z201", "Z303", "Z308", current_protocol]
 
     # This is pretty excruciating.  Details:
     #
@@ -374,6 +389,9 @@ class Connection(smac.SizedMessageAsyncConnection, object):
     #         sends Z303 to server
     #     OK, because Z303 is in the server's clients_we_can_talk_to
 
+    # Exception types that should not be logged:
+    unlogged_exception_types = ()
+
     # Client constructor passes 'C' for tag, server constructor 'S'.  This
     # is used in log messages, and to determine whether we can speak with
     # our peer.
@@ -403,16 +421,6 @@ class Connection(smac.SizedMessageAsyncConnection, object):
         ourmap = {}
         self.__super_init(sock, addr, map=ourmap)
 
-        # A Connection either uses asyncore directly or relies on an
-        # asyncore mainloop running in a separate thread.  If
-        # thr_async is true, then the mainloop is running in a
-        # separate thread.  If thr_async is true, then the asyncore
-        # trigger (self.trigger) is used to notify that thread of
-        # activity on the current thread.
-        self.thr_async = False
-        self.trigger = None
-        self._prepare_async()
-
         # The singleton dict is used in synchronous mode when a method
         # needs to call into asyncore to try to force some I/O to occur.
         # The singleton dict is a socket map containing only this object.
@@ -471,22 +479,17 @@ class Connection(smac.SizedMessageAsyncConnection, object):
         self.logger.log(level, self.log_label + message, exc_info=exc_info)
 
     def close(self):
+        self.mgr.close_conn(self)
         if self.closed:
             return
         self._singleton.clear()
         self.closed = True
         self.__super_close()
-        self.close_trigger()
+        self.trigger.pull_trigger()
         self.replies_cond.acquire()
         self.replies_cond.notifyAll()
         self.replies_cond.release()
 
-    def close_trigger(self):
-        # Overridden by ManagedClientConnection.
-        if self.trigger is not None:
-            self.trigger.pull_trigger()
-            self.trigger.close()
-
     def register_object(self, obj):
         """Register obj as the true object to invoke methods on."""
         self.obj = obj
@@ -556,7 +559,7 @@ class Connection(smac.SizedMessageAsyncConnection, object):
 
     def handle_request(self, msgid, flags, name, args):
         obj = self.obj
-        
+
         if name.startswith('_') or not hasattr(obj, name):
             if obj is None:
                 if __debug__:
@@ -564,7 +567,7 @@ class Connection(smac.SizedMessageAsyncConnection, object):
                              % (name, short_repr(args)),
                              level=logging.DEBUG)
                 return
-                
+
             msg = "Invalid method name: %s on %s" % (name, repr(obj))
             raise ZRPCError(msg)
         if __debug__:
@@ -581,8 +584,9 @@ class Connection(smac.SizedMessageAsyncConnection, object):
         except (SystemExit, KeyboardInterrupt):
             raise
         except Exception, msg:
-            self.log("%s() raised exception: %s" % (name, msg), logging.INFO,
-                     exc_info=True)
+            if not isinstance(msg, self.unlogged_exception_types):
+                self.log("%s() raised exception: %s" % (name, msg),
+                         logging.INFO, exc_info=True)
             error = sys.exc_info()[:2]
             return self.return_error(msgid, flags, *error)
 
@@ -707,10 +711,7 @@ class Connection(smac.SizedMessageAsyncConnection, object):
         if self.closed:
             raise DisconnectedError()
         msgid = self.send_call(method, args, 0)
-        if self.is_async():
-            self.trigger.pull_trigger()
-        else:
-            asyncore.poll(0.01, self._singleton)
+        self.trigger.pull_trigger()
         return msgid
 
     def _deferred_wait(self, msgid):
@@ -749,41 +750,13 @@ class Connection(smac.SizedMessageAsyncConnection, object):
         for method, args in iterator:
             yield self.__call_message(method, args, ASYNC)
 
-    # handle IO, possibly in async mode
-
-    def _prepare_async(self):
-        self.thr_async = False
-        ThreadedAsync.register_loop_callback(self.set_async)
-        # TODO:  If we are not in async mode, this will cause dead
-        # Connections to be leaked.
-
-    def set_async(self, map):
-        self.trigger = trigger()
-        self.thr_async = True
-
-    def is_async(self):
-        # Overridden by ManagedConnection
-        if self.thr_async:
-            return 1
-        else:
-            return 0
-
-    def _pull_trigger(self, tryagain=10):
-        try:
-            self.trigger.pull_trigger()
-        except OSError:
-            self.trigger.close()
-            self.trigger = trigger()
-            if tryagain > 0:
-                self._pull_trigger(tryagain=tryagain-1)
 
     def wait(self, msgid):
         """Invoke asyncore mainloop and wait for reply."""
         if __debug__:
-            self.log("wait(%d), async=%d" % (msgid, self.is_async()),
-                     level=TRACE)
-        if self.is_async():
-            self._pull_trigger()
+            self.log("wait(%d)" % msgid, level=TRACE)
+
+        self.trigger.pull_trigger()
 
         # Delay used when we call asyncore.poll() directly.
         # Start with a 1 msec delay, double until 1 sec.
@@ -801,7 +774,6 @@ class Connection(smac.SizedMessageAsyncConnection, object):
                         self.log("wait(%d): reply=%s" %
                                  (msgid, short_repr(reply)), level=TRACE)
                     return reply
-                assert self.is_async() # XXX we're such cowards
                 self.replies_cond.wait()
         finally:
             self.replies_cond.release()
@@ -816,84 +788,36 @@ class Connection(smac.SizedMessageAsyncConnection, object):
     def poll(self):
         """Invoke asyncore mainloop to get pending message out."""
         if __debug__:
-            self.log("poll(), async=%d" % self.is_async(), level=TRACE)
-        if self.is_async():
-            self._pull_trigger()
-        else:
-            asyncore.poll(0.0, self._singleton)
+            self.log("poll()", level=TRACE)
+        self.trigger.pull_trigger()
 
-    def _pending(self, timeout=0):
-        """Invoke mainloop until any pending messages are handled."""
-        if __debug__:
-            self.log("pending(), async=%d" % self.is_async(), level=TRACE)
-        if self.is_async():
-            return
-        # Inline the asyncore poll() function to know whether any input
-        # was actually read.  Repeat until no input is ready.
 
-        # Pending does reads and writes.  In the case of server
-        # startup, we may need to write out zeoVerify() messages.
-        # Always check for read status, but don't check for write status
-        # only there is output to do.  Only continue in this loop as
-        # long as there is data to read.
-        r = r_in = [self._fileno]
-        x_in = []
-        while r and not self.closed:
-            if self.writable():
-                w_in = [self._fileno]
-            else:
-                w_in = []
-            try:
-                r, w, x = select.select(r_in, w_in, x_in, timeout)
-            except select.error, err:
-                if err[0] == errno.EINTR:
-                    timeout = 0
-                    continue
-                else:
-                    raise
-            else:
-                # Make sure any subsequent select does not block.  The
-                # loop is only intended to make sure all incoming data is
-                # returned.
-
-                # Insecurity:  What if the server sends a lot of
-                # invalidations, such that pending never finishes?  Seems
-                # unlikely, but possible.
-                timeout = 0
-            if r:
-                try:
-                    self.handle_read_event()
-                except asyncore.ExitNow:
-                    raise
-                except:
-                    self.handle_error()
-            if w:
-                try:
-                    self.handle_write_event()
-                except asyncore.ExitNow:
-                    raise
-                except:
-                    self.handle_error()
 
 class ManagedServerConnection(Connection):
     """Server-side Connection subclass."""
-    __super_init = Connection.__init__
-    __super_close = Connection.close
+
+    # Exception types that should not be logged:
+    unlogged_exception_types = (ZODB.POSException.POSKeyError, )
+
+    # Servers use a shared server trigger that uses the asyncore socket map
+    trigger = trigger()
 
     def __init__(self, sock, addr, obj, mgr):
         self.mgr = mgr
-        self.__super_init(sock, addr, obj, 'S')
+        Connection.__init__(self, sock, addr, obj, 'S')
         self.marshal = ServerMarshaller()
-        self.obj.notifyConnected(self)
 
     def handshake(self):
         # Send the server's preferred protocol to the client.
         self.message_output(self.current_protocol)
 
+    def recv_handshake(self, proto):
+        Connection.recv_handshake(self, proto)
+        self.obj.notifyConnected(self)
+
     def close(self):
         self.obj.notifyDisconnected()
-        self.mgr.close_conn(self)
-        self.__super_close()
+        Connection.close(self)
 
 class ManagedClientConnection(Connection):
     """Client-side Connection subclass."""
@@ -901,6 +825,8 @@ class ManagedClientConnection(Connection):
     __super_close = Connection.close
     base_message_output = Connection.message_output
 
+    trigger = client_trigger
+
     def __init__(self, sock, addr, mgr):
         self.mgr = mgr
 
@@ -919,8 +845,6 @@ class ManagedClientConnection(Connection):
         self.queued_messages = []
 
         self.__super_init(sock, addr, None, tag='C', map=client_map)
-        self.thr_async = True
-        self.trigger = client_trigger
         client_trigger.pull_trigger()
 
     # Our message_ouput() queues messages until recv_handshake() gets the
@@ -964,27 +888,3 @@ class ManagedClientConnection(Connection):
             self.queue_output = False
         finally:
             self.output_lock.release()
-
-    # Defer the ThreadedAsync work to the manager.
-
-    def close_trigger(self):
-        # We are using a shared trigger for all client connections.
-        # We never want to close it.
-
-        # We do want to pull it to make sure the select loop detects that
-        # we're closed.
-        self.trigger.pull_trigger()
-
-    def set_async(self, map):
-        pass
-
-    def _prepare_async(self):
-        # Don't do the register_loop_callback that the superclass does
-        pass
-
-    def is_async(self):
-        return True
-
-    def close(self):
-        self.mgr.close_conn(self)
-        self.__super_close()
